@@ -23,7 +23,11 @@ pub enum Error {
     /// There was not enough data present to parse.
     MissingData,
     /// A boolean had a value other than true or false.
-    InvalidBool
+    InvalidBool,
+    /// While reading NBT, the stream started with a value other than 0x0a.
+    InvalidNBTHeader,
+    /// While reading NBT, the stream had an invalid data type ID.
+    InvalidNBTType
 }
 
 impl std::fmt::Display for Error {
@@ -34,6 +38,162 @@ impl std::fmt::Display for Error {
 
 impl std::error::Error for Error {}
 
+/// Provides tools for reading and managing NBT types.
+pub mod nbt {
+    use super::{Error, read_byte};
+    /// Reads an entire NBT compound from a Read type.
+    pub fn from_reader<R: std::io::Read>(reader: &mut R) -> Result<NamedTag, Error> {
+        if read_byte(reader)? != 0x0a {
+            return Err(Error::InvalidNBTHeader);
+        }
+        let root_name = named_tag_name_reader(reader)?;
+        let mut elements = vec![];
+        loop {
+            let next_tag = read_named_tag(reader)?;
+            match next_tag.tag {
+                Tag::End => {
+                    break;
+                }
+                _ => {
+                    elements.push(next_tag);
+                }
+            }
+        }
+        return Ok(NamedTag { name: root_name, tag: Tag::Compound(elements) });
+    }
+    fn named_tag_name_reader<R: std::io::Read>(reader: &mut R) -> Result<String, Error> {
+        let string_len = u16::from_be_bytes([read_byte(reader)?; 2]);
+        let mut bytes = vec![];
+        for _ in 0..string_len {
+            bytes.push(read_byte(reader)?);
+        }
+        // This is required because Mojang uses Java's modified utf-8 which isn't supported here
+        unsafe {
+            let string = String::from_utf8_unchecked(bytes);
+            return Ok(string);
+        }
+    }
+    fn read_named_tag<R: std::io::Read>(reader: &mut R) -> Result<NamedTag, Error> {
+        let tag_type = read_byte(reader)?;
+        let tag_name;
+        if !(tag_type == 0x00) {
+            tag_name = named_tag_name_reader(reader)?;
+        }
+        else {
+            tag_name = String::from("N/A");
+        }
+        let tag_val = read_from_type(reader, tag_type)?;
+        return Ok(NamedTag { name: tag_name, tag: tag_val });
+    }
+    fn read_tag<R: std::io::Read>(reader: &mut R) -> Result<Tag, Error> {
+        let tag_type = read_byte(reader)?;
+        let tag_val = read_from_type(reader, tag_type)?;
+        return Ok(tag_val);
+    }
+    fn read_from_type<R: std::io::Read>(reader: &mut R, type_id: u8) -> Result<Tag, Error> {
+        match type_id {
+            0x00 => {
+                return Ok(Tag::End);
+            }
+            0x01 => {
+                return Ok(Tag::Byte(i8::from_be_bytes([read_byte(reader)?])));
+            }
+            0x02 => {
+                return Ok(Tag::Short(i16::from_be_bytes([read_byte(reader)?; 2])));
+            }
+            0x03 => {
+                return Ok(Tag::Int(i32::from_be_bytes([read_byte(reader)?; 4])));
+            }
+            0x04 => {
+                return Ok(Tag::Long(i64::from_be_bytes([read_byte(reader)?; 8])));
+            }
+            0x05 => {
+                return Ok(Tag::Float(f32::from_be_bytes([read_byte(reader)?; 4])));
+            }
+            0x06 => {
+                return Ok(Tag::Double(f64::from_be_bytes([read_byte(reader)?; 8])));
+            }
+            0x07 => {
+                let array_len = i32::from_be_bytes([read_byte(reader)?; 4]);
+                let mut array = vec![];
+                for _ in 0..array_len {
+                    array.push(i8::from_be_bytes([read_byte(reader)?]));
+                }
+                return Ok(Tag::ByteArray(array));
+            }
+            0x08 => {
+                return Ok(Tag::String(named_tag_name_reader(reader)?));
+            }
+            0x09 => {
+                let _list_type = read_byte(reader)?;
+                let list_len = i32::from_be_bytes([read_byte(reader)?; 4]);
+                if list_len < 1 {
+                    return Ok(Tag::List(vec![Tag::End]));
+                }
+                let mut list_elements = vec![];
+                for _ in 0..list_len {
+                    list_elements.push(read_tag(reader)?);
+                }
+                return Ok(Tag::List(list_elements))
+            }
+            0x0A => {
+                let mut compound_elements = vec![];
+                loop {
+                    let tag = read_named_tag(reader)?;
+                    if tag.tag == Tag::End {
+                        break;
+                    }
+                    else {
+                        compound_elements.push(tag);
+                    }
+                }
+                return Ok(Tag::Compound(compound_elements));
+            }
+            0x0B => {
+                let array_len = i32::from_be_bytes([read_byte(reader)?; 4]);
+                let mut array = vec![];
+                for _ in 0..array_len {
+                    array.push(i32::from_be_bytes([read_byte(reader)?; 4]));
+                }
+                return Ok(Tag::IntArray(array));
+            }
+            0x0C => {
+                let array_len = i32::from_be_bytes([read_byte(reader)?; 4]);
+                let mut array = vec![];
+                for _ in 0..array_len {
+                    array.push(i64::from_be_bytes([read_byte(reader)?; 8]));
+                }
+                return Ok(Tag::LongArray(array));
+            }
+            _ => {
+                return Err(Error::InvalidNBTType);
+            }
+        }
+    }
+    #[derive(PartialEq, Clone, Debug)]
+    pub enum Tag {
+        Byte(i8),
+        Short(i16),
+        Int(i32),
+        Long(i64),
+        Float(f32),
+        Double(f64),
+        ByteArray(Vec<i8>),
+        String(String),
+        List(Vec<Tag>),
+        Compound(Vec<NamedTag>),
+        IntArray(Vec<i32>),
+        LongArray(Vec<i64>),
+        End
+    }
+
+    #[derive(PartialEq, Clone, Debug)]
+    pub struct NamedTag {
+        pub name: String,
+        pub tag: Tag
+    }
+}
+
 /// Represents a Java Int (i32) using between 1-5 bytes.
 #[derive(Eq, Clone, Copy, Debug)]
 pub struct VarInt {
@@ -828,9 +988,9 @@ fn read_byte<R: std::io::Read>(reader: &mut R) -> Result<u8, Error> {
 }
 
 mod test {
-    use super::*;
     #[test]
-    fn varint_standard_values() -> Result<(), Error> {
+    fn varint_standard_values() -> Result<(), super::Error> {
+        use super::VarInt;
         // Create the list of standard values
         let val_0 = VarInt::from_value(0)?;
         let val_1 = VarInt::from_value(1)?;
@@ -861,7 +1021,8 @@ mod test {
         return Ok(());
     }
     #[test]
-    fn varlong_standard_values() -> Result<(), Error> {
+    fn varlong_standard_values() -> Result<(), super::Error> {
+        use super::VarLong;
         // Create the list of standard values
         let val_0 = VarLong::from_value(0)?;
         let val_1 = VarLong::from_value(1)?;
@@ -885,7 +1046,8 @@ mod test {
         return Ok(());
     }
     #[test]
-    fn position_standard_values() -> Result<(), Error> {
+    fn position_standard_values() -> Result<(), super::Error> {
+        use super::Position;
         // Create the list of standard values
         let zeroed = Position::from_values(0, 0, 0);
         let max_value = Position::from_values(i32::MAX, i16::MAX, i32::MAX);
